// Lz4Decode:
#ifndef FLAT32
//   input   ds:si=inStream, es:di=outStream
//   output  outStream[], ds:si, es:di
	.code16
#define AX	%ax
#define BX	%bx
#define CX	%cx
#define SI	%si
#define DI	%di
#else
//   input   esi=inStream, edi=outStream
//   output  outStream[], ds:esi, es:edi
	.code32
#define AX	%eax
#define BX	%ebx
#define CX	%ecx
#define SI	%esi
#define DI	%edi
#endif
 
#if defined(FLAT16OUT) || defined(ONLY8086)
#define cAX	%ax
#define cCX	%cx
#define cDX	%dx
#else
#define cAX	%eax
#define cCX	%ecx
#define cDX	%edx
#endif

#define ARCHIVE_MAGICNUMBER 0x184C2102

//#define PARANOIA			// cover rare cases, optional

lz4main:
#ifdef PARANOIA
	cld
# if !defined(FLAT32) && !defined(FLAT16OUT)
	xorl	%cx, %cx
	call	lz4mov
# endif
#endif
#if defined(FLAT16OUT) || defined(ONLY8086)
	lodsw			// get chunkSize
# ifndef NO_LZ4_HEADER
	cmpw	$ARCHIVE_MAGICNUMBER&0xFFFF, %ax
# endif
	xchg	%ax, %dx
	lodsw
# ifndef NO_LZ4_HEADER
	jne	chkeof
	cmpw	$ARCHIVE_MAGICNUMBER>>16, %ax
	je	lz4main
chkeof:	
# endif
# ifdef PARANOIA
	orw	%ax, %dx	// end of file ?
	je	lz4quit
# endif
# ifndef FLAT16OUT
	xchgw	%ax, %bp
	subw	$1, %dx
	sbbw	$0, %bp
# endif
#else
	lodsl			// get chunkSize
# ifndef NO_LZ4_HEADER
	cmpl	$ARCHIVE_MAGICNUMBER, %eax
	je	lz4main
# endif
# ifdef PARANOIA
	orl	%eax, %eax	// end of file ?
	je	lz4quit
# endif
	xchgl	%eax, %edx
#endif
lz4chunk:			// uncompress chunk
	lodsb			// get token
	pushw	%ax
#ifdef ONLY8086
	movb	$4, %cl
	shrb	%cl, %al
#else
	shrb	$4, %al
#endif
	call	lz4len		// get literal length
#if !defined(FLAT16OUT) && defined(ONLY8086)
	subw	%cx, %dx	// count literal
	sbbw	$0, %bp
#else
	sub	cCX, cDX	// count literal
#endif
#if !defined(FLAT32) && !defined(FLAT16OUT) && (!defined(FLAT16) || !defined(PARANOIA))
#define NeedLz4mov
	call	lz4mov		// copy literals
#else
	rep movsb
#endif
	popw	%bx
#if !defined(FLAT16OUT) && defined(ONLY8086)
	subw	$1+2, %dx	// count token & string address
	sbbw	$0, %bp
	jb	lz4quit
lz4cont:
#else
	sub	$1+2, cDX	// count token & string address
	jbe	lz4quit
#endif
	lodsw			// get string address
	xchg	AX, BX
	call	lz4len		// get string length
	add	$4, CX
#if !defined(FLAT32) && !defined(FLAT16OUT)
	pushw	%ds
	pushw	%si
	movw	%di, %si
	subw	%bx, %si
	movw	%es, %ax
	jnc	axok
	subb	$0x10, %ah
axok:
.macro norm	reg
	movw	%si, \reg
	andw	$0xF, %si
# ifdef ONLY8086
	pushw	%cx
	movb	$4, %cl
	shrw	%cl, \reg
# else
	shrw	$4, \reg
# endif
	addw	\reg, %ax
	movw	%ax, %ds
	movw	%di, \reg
	andw	$0xF, %di
# ifdef ONLY8086
	shrw	%cl, \reg
	popw	%cx
# else
	shrw	$4, \reg
# endif
	movw	%es, %ax
	addw	\reg, %ax
	movw	%ax, %es
.endm 
# if !defined (NeedLz4mov)
	norm	%bx
	rep movsb
# else
	call	lz4movStr	// copy string
# endif
	popw	%si
	popw	%ds
#else
# ifdef FASTFILL
	cmp	$FASTFILL,BX
	jbe	lz4fast
# endif
lz4notfast:
	xchg	AX, SI
	mov	DI, SI
	sub	BX, SI
# ifdef ONLY8086
lz4movsb:
	movsb	%es:(SI), %es:(DI)	//  NMOS 8088/8086 workaround.
	loop	lz4movsb
# else
	rep movsb %es:(SI), %es:(DI)
# endif
	xchg	AX, SI
# ifdef FASTFILL
lz4fast:
#  if FASTFILL == 1
	movb	%es:-1(DI), %al
	rep stosb
#  endif
#  if FASTFILL >= 2
	movw	%es:-2(DI), %ax
	je	lz4fastword
	movb	%al, %ah
lz4fastword:
	shr	$1, CX
	rep stosw
	jnc	lz4chunk
	stosb
#  endif
# endif
	jmp	lz4chunk
#endif

#if defined(NeedLz4mov)
# if defined(PARANOIA)
lz4movlp:
	xchgw	%ax, %cx
	movw	$0x7800, %cx
	rep movsw
	xchgw	%ax, %cx
	sub	$0xF0, %ch
# endif
lz4mov:
	movw	%ds, %ax
lz4movStr:
	norm	%bx
# if defined(PARANOIA)
	cmp	$0xFF, %ch	// catch FFFX case
	jz	lz4movlp
# endif
	rep movsb
	ret
#endif

lz4len:				// get length in %ecx
	and	$0xF, cAX
	mov	cAX, cCX
	cmpb	$0xF, %al
	jne	lz4quit
lz4len2:
	lodsb
#if !defined(FLAT16OUT) && defined(ONLY8086)
	subw	$1, %dx		// remaining chunk size
	sbbw	$0, %bp
#else
	dec	cDX		// remaining chunk size
#endif
	add	AX, CX
	cmpb	$0xFF, %al
	je	lz4len2
lz4quit:
	ret
